import django

__all__ = ['User', 'get_user_model']


# This small file just keeps the app compatible with Django versions 1.5+ and
# < 1.5. There are some "unused" imports like `get_user_model` or `User` but
# they are ther to get imported like `from .compat import get_user_model`

if django.VERSION >= (1, 5):
    from django.conf import settings
    from django.contrib.auth import get_user_model
    User = settings.AUTH_USER_MODEL if hasattr(settings, 'AUTH_USER_MODEL') else 'auth.User'

else:
    from django.contrib.auth.models import User

    def get_user_model():
        return User



from optparse import make_option

from django.core.management.base import BaseCommand
from django.utils.six.moves import input

from hashbrown.models import Switch
from hashbrown.utils import SETTINGS_KEY, is_active, get_defaults


class Command(BaseCommand):
    help = 'Creates / deletes feature switches in the database'

    option_list = BaseCommand.option_list + (
        make_option(
            '--delete',
            action='store_true',
            default=False,
            help='Delete switches in the database that are not in ' + SETTINGS_KEY,
        ),
        make_option(
            '--force',
            action='store_true',
            default=False,
            help='Delete switches without confirmation (implies --delete)',
        ),
    )

    def handle(self, *args, **kwargs):
        if kwargs['delete'] or kwargs['force']:
            self._delete_switches(force=kwargs['force'])

        self._create_switches()
        self.stderr.write('All switches up-to-date.')

    def _create_switches(self):
        create_switches(self.stderr)

    def _delete_switches(self, force=False):
        delete_switches(self.stderr, force=force)


def create_switches(stderr):
    """Create switches listed in HASHBROWN_SWITCH_DEFAULTS which aren't in
    the database yet.
    """
    defaults = get_defaults()
    installed_switches = set(Switch.objects.values_list('label', flat=True))
    missing_switches = set(defaults) - installed_switches

    for label in sorted(missing_switches):
        is_active(label)
        stderr.write('Created switch %r.' % label)

    return missing_switches


def delete_switches(stderr, force=False):
    defaults = get_defaults()
    installed_switches = set(Switch.objects.values_list('label', flat=True))
    unknown_switches = sorted(installed_switches - set(defaults))

    if not unknown_switches:
        return

    permission_granted = force or ask_permission(stderr, unknown_switches)

    if permission_granted:
        Switch.objects.filter(label__in=unknown_switches).delete()

        for label in unknown_switches:
            stderr.write('Deleted switch %r.' % label)


def ask_permission(stderr, switches):
    stderr.write('The following switches are in the database but not in %s:' % SETTINGS_KEY)

    for label in switches:
        stderr.write(label)

    response = input('Delete switches? [y/N]: ')

    return response.lower().strip() in ('y', 'yes')



from django.db import models

from .compat import User


class Switch(models.Model):

    label = models.CharField(max_length=200, unique=True)
    description = models.TextField(
        help_text='Short description of what this switch is doing', blank=True)
    globally_active = models.BooleanField(default=False)

    users = models.ManyToManyField(
        User, null=True, related_name='available_switches', blank=True)

    class Meta:
        verbose_name_plural = u'switches'

    def __unicode__(self):
        return self.label



from functools import wraps

import hashbrown


class SwitchesContext(object):
    def __init__(self, **kwargs):
        self.keys = kwargs
        self.is_active_func = hashbrown.is_active

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return inner

    def __enter__(self):
        self.patch()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unpatch()

    def patch(self):
        def is_active_patched(tag):
            return self.keys.get(tag, self.is_active_func(tag))

        hashbrown.is_active = is_active_patched

    def unpatch(self):
        hashbrown.is_active = self.is_active_func


switches = SwitchesContext



from django.contrib import admin

from .models import Switch


class SwitchAdmin(admin.ModelAdmin):
    list_display = ('label', 'globally_active', 'description')
    list_editable = ('globally_active',)


admin.site.register(Switch, SwitchAdmin)



from django.core.management import call_command
from django.db import IntegrityError
from django.template import Context, Template, TemplateSyntaxError
from django.test import TestCase
from django.test.utils import override_settings

import hashbrown
from .compat import get_user_model
from .models import Switch
from .testutils import switches


HASHBROWN_SWITCH_DEFAULTS = {
    'test': {
        'globally_active': True
    },
    'things': {
        'globally_active': False,
        'description': 'This does some things'
    }
}


class UtilsTestCase(TestCase):

    def test_is_active_without_existing_flag_creates_it(self):
        self.assertFalse(Switch.objects.filter(label='some_feature').exists())

        self.assertFalse(hashbrown.is_active('some_feature'))

        self.assertTrue(Switch.objects.filter(label='some_feature').exists())

    def test_is_active_with_existing_disabled_flag(self):
        Switch.objects.create(label='some_feature', globally_active=False)

        with self.assertNumQueries(1):  # get
            self.assertFalse(hashbrown.is_active('some_feature'))

        self.assertEqual(
            Switch.objects.filter(label='some_feature').count(), 1)

    def test_is_active_with_existing_enabled_flag(self):
        Switch.objects.create(label='some_feature', globally_active=True)

        with self.assertNumQueries(1):  # get
            self.assertTrue(hashbrown.is_active('some_feature'))

        self.assertEqual(
            Switch.objects.filter(label='some_feature').count(), 1)

    def test_is_active_disabled_globally_for_users(self):
        user = get_user_model().objects.create(
            email='test@example.com', username='test')

        Switch.objects.create(label='some_feature', globally_active=False)

        with self.assertNumQueries(2):  # get, check users
            self.assertFalse(hashbrown.is_active('some_feature', user=user))

    def test_is_active_enabled_globally_for_users(self):
        user = get_user_model().objects.create(
            email='test@example.com', username='test')

        Switch.objects.create(label='some_feature', globally_active=True)

        with self.assertNumQueries(1):  # get
            self.assertTrue(hashbrown.is_active('some_feature', user=user))

    def test_is_active_for_certain_user_with_flag_enabled(self):
        user = get_user_model().objects.create(
            email='test@example.com', username='test')
        switch = Switch.objects.create(
            label='some_feature', globally_active=True)
        switch.users.add(user)

        with self.assertNumQueries(1):  # get
            self.assertTrue(hashbrown.is_active('some_feature', user=user))

    def test_is_active_for_different_user_with_flag_enabled(self):
        user_1 = get_user_model().objects.create(
            email='test@example.com', username='test')
        user_2 = get_user_model().objects.create(
            email='test@example.com', username='test2')
        switch = Switch.objects.create(
            label='some_feature', globally_active=False)
        switch.users.add(user_1)

        with self.assertNumQueries(2):  # get, check user
            self.assertFalse(hashbrown.is_active('some_feature', user=user_2))

    @override_settings(HASHBROWN_SWITCH_DEFAULTS=HASHBROWN_SWITCH_DEFAULTS)
    def test_default_switches_on_settings(self):
        self.assertTrue(hashbrown.is_active('test'))

        self.assertFalse(hashbrown.is_active('things'))

        self.assertEqual(
            Switch.objects.get(label='things').description,
            'This does some things'
        )


class TemplateTagsTestCase(TestCase):
    def test_simple(self):
        Switch.objects.create(label='test', globally_active=True)

        template = Template("""
            {% load hashbrown_tags %}
            {% ifswitch 'test' %}
            hello world!
            {% endifswitch %}
        """)
        rendered = template.render(Context())

        self.assertTrue('hello world!' in rendered)

    def test_simple_new_switch(self):
        template = Template("""
            {% load hashbrown_tags %}
            {% ifswitch 'test' %}
            hello world!
            {% endifswitch %}
        """)
        rendered = template.render(Context())

        self.assertFalse('hello world!' in rendered)
        self.assertTrue(Switch.objects.filter(label='test').exists())

    def test_not_closing_raises_error(self):
        self.assertRaises(TemplateSyntaxError, Template, """
            {% load hashbrown_tags %}
            {% ifswitch 'test' %}
            hello world!
        """)

    def test_no_attribute_raises_error(self):
        self.assertRaises(TemplateSyntaxError, Template, """
            {% load hashbrown_tags %}
            {% ifswitch %}
            hello world!
            {% endifswitch %}
        """)

    def test_else(self):
        template = Template("""
            {% load hashbrown_tags %}
            {% ifswitch 'test' %}
            hello world!
            {% else %}
            things!
            {% endifswitch %}
        """)
        rendered = template.render(Context())

        self.assertFalse('hello world!' in rendered)
        self.assertTrue('things!' in rendered)

    def test_with_user(self):
        user_1 = get_user_model().objects.create(
            email='test@example.com', username='test')
        user_2 = get_user_model().objects.create(
            email='test@example.com', username='test2')
        switch = Switch.objects.create(
            label='some_feature', globally_active=False)
        switch.users.add(user_1)

        template = Template("""
            {% load hashbrown_tags %}
            {% ifswitch 'some_feature' user %}
            hello world!
            {% else %}
            things!
            {% endifswitch %}
        """)

        rendered = template.render(Context({'user': user_1}))
        self.assertTrue('hello world!' in rendered)
        self.assertFalse('things!' in rendered)

        rendered = template.render(Context({'user': user_2}))
        self.assertFalse('hello world!' in rendered)
        self.assertTrue('things!' in rendered)


class TestUtilsTestCase(TestCase):

    def test_as_decorator_active(self):
        @switches(things=True)
        def test():
            return hashbrown.is_active('things')

        self.assertTrue(test())

    def test_as_decorator_inactive(self):
        @switches(things=False)
        def test():
            return hashbrown.is_active('things')

        self.assertFalse(test())


class SwitchModelTestCase(TestCase):
    def test_label_is_unique(self):
        foo_switch = Switch.objects.create(label='foo')

        with self.assertRaises(IntegrityError):
            Switch.objects.create(label='foo')


class ManagementCommandTestCase(TestCase):
    def test_switches_command_creates_missing_switches(self):
        self.assertEqual(Switch.objects.count(), 0)

        with self.settings(HASHBROWN_SWITCH_DEFAULTS=HASHBROWN_SWITCH_DEFAULTS):
            call_command('switches')

        self.assertItemsEqual(
            Switch.objects.values_list('label', flat=True),
            ['test', 'things'],
        )

    def test_switches_command_ignores_existing_switches(self):
        foo = Switch.objects.create(
            label='foo',
            globally_active=False,
            description='Before description',
        )

        defaults = {
            'foo': {
                'globally_active': True,
                'description': 'After description',
            },
        }

        with self.settings(HASHBROWN_SWITCH_DEFAULTS=defaults):
            call_command('switches')

        self.assertItemsEqual(
            Switch.objects.values(),
            [
                {
                    'id': foo.pk,
                    'label': 'foo',
                    'globally_active': False,
                    'description': 'Before description',
                }
            ],
        )

    def test_force_flag_deletes_unknown_switches(self):
        Switch.objects.create(label='foo')

        with self.settings(HASHBROWN_SWITCH_DEFAULTS={}):
            call_command('switches', force=True)

        self.assertEqual(Switch.objects.count(), 0)



from django.conf import settings
from .models import Switch


SETTINGS_KEY = 'HASHBROWN_SWITCH_DEFAULTS'


def is_active(label, user=None):
    defaults = get_defaults()

    globally_active = defaults[label].get(
        'globally_active',
        False) if label in defaults else False

    description = defaults[label].get(
        'description',
        '') if label in defaults else ''

    switch, created = Switch.objects.get_or_create(
        label=label, defaults={
            'globally_active': globally_active,
            'description': description,
        })

    if created:
        return switch.globally_active

    if switch.globally_active or (
        user and user.available_switches.filter(pk=switch.pk).exists()
    ):
        return True
    return False


def get_defaults():
    return getattr(settings, SETTINGS_KEY, {})



import os
from setuptools import setup, find_packages


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


setup(
    name="django-hashbrown",
    version="0.6",
    author="Pablo Recio",
    author_email="pablo@potatolondon.com",
    description="Yet another dead simple feature switching library for Django.",
    long_description=(read('README.md')),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    license="BSD",
    keywords="django feature switching potato",
    url='https://github.com/potatolondon/django-hashbrown',
    packages=find_packages(),
    zip_safe=False,
)



from django import template
from ..utils import is_active

register = template.Library()


@register.tag
def ifswitch(parser, token):
    bits = token.split_contents()

    if len(bits) < 2:
        raise template.TemplateSyntaxError(
            "%r tag requires an argument" % bits[0])

    name = bits[1]

    try:
        user = bits[2]
    except IndexError:
        user = None

    nodelist_true = parser.parse(('else', 'endifswitch'))
    token = parser.next_token()

    if token.contents == 'else':
        nodelist_false = parser.parse(('endifswitch',))
        parser.delete_first_token()
    else:
        nodelist_false = template.NodeList()

    return SwitchNode(nodelist_true, nodelist_false, name, user)


class SwitchNode(template.Node):
    def __init__(self, nodelist_true, nodelist_false, name, user):
        self.nodelist_true = nodelist_true
        self.nodelist_false = nodelist_false
        self.name = template.Variable(name)

        if user:
            self.user = template.Variable(user)
        else:
            self.user = None

    def render(self, context):
        if self.user:
            user = self.user.resolve(context)
        else:
            user = None

        if not is_active(self.name.resolve(context), user):
            return self.nodelist_false.render(context)

        return self.nodelist_true.render(context)
